#ifndef __ALGORITHM_H
#define __ALGORITHM_H

#include <iostream>

#include "chromosome.h"
#include "generation.h"
#include "fitness/fitness.h"
#include "generator/generator.h"
#include "selection/selection.h"
#include "crossover/crossover.h"
#include "mutation/mutation.h"

#include "condition/condition.h"

using namespace std;

namespace genetic {
    /**
     * Genetic Algorithm itself
     */
    template < typename _Chromosome, typename _Selection, typename _Crossover, typename _Mutation, typename _Fitness = Fitness<_Chromosome, typename _Selection::FitnessValueType > >
    class Algorithm {
    public:
        /**
         * Type representing Fitness value
         */
        typedef typename _Selection::FitnessValueType FitnessValueType;
    protected:
        /**
         * Generator which draws initial population
         */
        generator::Generator<_Chromosome>& generator;

        /**
         * Fitness function used in selection
         *
         * It is a reference, because Fitness is purely virtual class and we
         * can't have instance of it. It must be one of derived classes.
         */
        _Fitness& fitness;

        /**
         * Crossover function
         */
        _Crossover crossover;

        /**
         * Mutation function
         */
        _Mutation mutation;

        /**
         * Current population
         */
        Generation<_Chromosome> generation;

        /**
         * Method invoked in each iteration of algorithm.
         *
         * Applies selection, crossover and mutation on each of the chromosomes
         * in generation.
         */
        void do_search() {
            _Selection selection(this->generation, fitness);
            this->generation = selection.draw();
            this->generation = crossover.cross(this->generation);
            this->generation = mutation.mutate(this->generation);
        }
    public:
        /**
         * Class constructor. Initializes class with values and breeds initial
         * population using passed generator::Generation.
         *
         * @param _generator Generator used to generate initial population
         * @param _fitness Fitness class used to check fitness of the chromosome
         * @param crossoverChance probability of crossover (0 = 0%, 1 = 100%)
         * @param mutationChance probability of fitness (0 = 0%, 1 = 100%)
         */
        Algorithm(
            generator::Generator<_Chromosome>& _generator,
            _Fitness& _fitness,
            float crossoverChance,
            float mutationChance) :
                generator(_generator),
                fitness(_fitness),
                crossover(crossoverChance),
                mutation(mutationChance) {

            this->generation = this->generator.breed();
        }

        /**
         * Starts Algorithm in search for result.
         * For each generation but initial, condition is checked, if algorithm
         * should stop.
         *
         * Displays separated by space: generation number, generation average
         * fitness and best chromosome fitness.
         *
         * @param condition Condition used to check after breeding new generation
         */
        void searchForResult(Condition<_Chromosome>& condition) {
            unsigned int i = 0;
            cout << "generation avg best\n";
            do {
                cout << i;
                do_search();

//                 cout << "New Generation:\n";
//                 this->showGeneration();
                this->showAvgFitness();
                this->showBestFitness();
            } while(condition.check(this->generation) && ++i);
        }

        /**
         * Displays entire generation.
         * \attention Method is currently not in use.
         */
        void showGeneration() {
            const unsigned int generationSize = this->generation.size();
            const unsigned int chromosomeSize = this->generation[0].size();

            for (unsigned int i = 0; i < generationSize; i++) {
                cout << "# " << i << ") ";
                for (unsigned int j = 0; j < chromosomeSize; j++) {
                    cout << this->generation[i][j].get();
                }
                cout << "\n";
            }
        }

        /**
         * Displays average fitness value of the entire generation.
         */
        void showAvgFitness() {
            float avg = 0;
            const unsigned int generationSize = this->generation.size();
            std::map<string, string> arguments = fitness.getArguments();

            for (unsigned int i = 0; i < generationSize; i++) {
                _Fitness fit(this->generation[i]);
                fit.parseArguments(arguments);
                avg += fit.calculate();
            }
            cout << " " << avg / generationSize;
        }

        /**
         * Displays best fitness value of the entire generation.
         */
        void showBestFitness() {
            float best = -100000;
            const unsigned int generationSize = this->generation.size();
            std::map<string, string> arguments = fitness.getArguments();

            for (unsigned int i = 0; i < generationSize; i++) {
                _Fitness fit(this->generation[i]);
                fit.parseArguments(arguments);

                float tmp = fit.calculate();
                if (tmp > best) {
                    best = tmp;
                }
            }
            cout << " " << best << "\n";
        }
    };
}

#endif /* __ALGORITHM_H */
